# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import pprint
import textwrap
from types import ModuleType
from typing import Any, Dict, Iterable, NamedTuple, Optional, Tuple, Type, Union

# import scripts.clientAPI as csgenome
# from gitlab import types as g_types
# from gitlab.exceptions import GitlabParsingError

# from .client import Gitlab, GitlabList
from .client import CSGenome
from .utils import handle_client_exception

__all__ = [
    "RequiredOptional",
    "RESTObject",
    "RESTObjectList",
    "RESTManager",
]


# _URL_ATTRIBUTE_ERROR = (
#     f"https://python-gitlab.readthedocs.io/en/v{gitlab.__version__}/"
#     f"faq.html#attribute-error-list"
# )


class RESTObject(object):
    """Represents an object built from server data.

    It holds the attributes know from the server, and the updated attributes in
    another. This allows smart updates, if the object allows it.

    You can redefine ``_id_attr`` in child classes to specify which attribute
    must be used as uniq ID. ``None`` means that the object can be updated
    without ID in the url.
    """

    _id_attr = "id"
    _attrs: Dict[str, Any]
    _created_from_list: bool  # Indicates if object was created from a list() action
    _module: ModuleType
    _parent_attrs: Dict[str, Any]
    _short_print_attr: Optional[str] = None
    _updated_attrs: Dict[str, Any]
    manager: "RESTManager"

    def __init__(self, manager, attrs, *, created_from_list=False):
        self.__dict__.update(
            {
                "manager": manager,
                "_attrs": attrs,
                "_updated_attrs": {},
                "_module": importlib.import_module(self.__module__),
                "_created_from_list": created_from_list,
            }
        )
        self.__dict__["_parent_attrs"] = self.manager.get_parent_attrs
        self._create_managers()

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()
        module = state.pop("_module")
        state["_module_name"] = module.__name__
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        module_name = state.pop("_module_name")
        self.__dict__.update(state)
        self.__dict__["_module"] = importlib.import_module(module_name)

    def __getattr__(self, name):
        try:
            return self.__dict__["_updated_attrs"][name]
        except KeyError:
            try:
                value = self.__dict__["_attrs"][name]

                # If the value is a list, we copy it in the _updated_attrs dict
                # because we are not able to detect changes made on the object
                # (append, insert, pop, ...). Without forcing the attr
                # creation __setattr__ is never called, the list never ends up
                # in the _updated_attrs dict, and the update() and save()
                # method never push the new data to the server.
                # See https://github.com/python-gitlab/python-gitlab/issues/306
                #
                # note: _parent_attrs will only store simple values (int) so we
                # don't make this check in the next except block.
                if isinstance(value, list):
                    self.__dict__["_updated_attrs"][name] = value[:]
                    return self.__dict__["_updated_attrs"][name]

                return value

            except KeyError:
                try:
                    return self.__dict__["_parent_attrs"][name]
                except KeyError as exc:
                    message = (
                        f"{type(self).__name__!r} object has no attribute {name!r}"
                    )
                    # if self._created_from_list:
                    #     message = (
                    #         f"{message}\n\n"
                    #         + textwrap.fill(
                    #             f"{self.__class__!r} was created via a list() call and "
                    #             f"only a subset of the data may be present. To ensure "
                    #             f"all data is present get the object using a "
                    #             f"get(object.id) call. For more details, see:"
                    #         )
                    #         + f"\n\n{_URL_ATTRIBUTE_ERROR}"
                    #     )
                    raise AttributeError(message) from exc

    def __setattr__(self, name, value):
        self.__dict__["_updated_attrs"][name] = value

    def __str__(self):
        data = self._attrs.copy()
        data.update(self._updated_attrs)
        return f"{type(self)} => {data}"

    def pformat(self) -> str:
        data = self._attrs.copy()
        data.update(self._updated_attrs)
        # return f"{type(self)} => \n{pprint.pformat(data)}"
        return f"{pprint.pformat(data)}"

    def pprint(self):
        print(self.pformat())

    def __repr__(self) -> str:
        if self._id_attr:
            return f"<{self.__class__.__name__} {self._id_attr}:{self.get_id()}>"
        else:
            return f"<{self.__class__.__name__}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RESTObject):
            return NotImplemented
        if self.get_id() and other.get_id():
            return self.get_id() == other.get_id()
        return super(RESTObject, self) == other

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, RESTObject):
            return NotImplemented
        if self.get_id() and other.get_id():
            return self.get_id() != other.get_id()
        return super(RESTObject, self) != other

    def __dir__(self) -> Iterable[str]:
        return set(self.attributes).union(super(RESTObject, self).__dir__())

    def __hash__(self) -> int:
        if not self.get_id():
            return super(RESTObject, self).__hash__()
        return hash(self.get_id())

    def _create_managers(self) -> None:
        # NOTE(jlvillal): We are creating our managers by looking at the class
        # annotations. If an attribute is annotated as being a *Manager type
        # then we create the manager and assign it to the attribute.
        for attr, annotation in sorted(self.__annotations__.items()):
            # We ignore creating a manager for the 'manager' attribute as that
            # is done in the self.__init__() method
            if attr in ("manager",):
                continue
            if not isinstance(annotation, (type, str)):
                continue
            if isinstance(annotation, type):
                cls_name = annotation.__name__
            else:
                cls_name = annotation
            # All *Manager classes are used except for the base "RESTManager" class
            if cls_name == "RESTManager" or not cls_name.endswith("Manager"):
                continue
            cls = getattr(self._module, cls_name)
            manager = cls(self.manager.gitlab, parent=self)
            # Since we have our own __setattr__ method, we can't use setattr()
            self.__dict__[attr] = manager

    def _update_attrs(self, new_attrs: Dict[str, Any]) -> None:
        self.__dict__["_updated_attrs"] = {}
        self.__dict__["_attrs"] = new_attrs

    def get_id(self) -> Optional[Union[int, str]]:
        """Returns the id of the resource."""
        if self._id_attr is None or not hasattr(self, self._id_attr):
            return None
        return getattr(self, self._id_attr)

    @property
    def encoded_id(self) -> Optional[Union[int, str]]:
        """Ensure that the ID is url-encoded so that it can be safely used in a URL
        path"""
        obj_id = self.get_id()
        # if isinstance(obj_id, str):
        #     obj_id = gitlab.utils.EncodedId(obj_id)
        return obj_id

    @property
    def attributes(self) -> Dict[str, Any]:
        d = self.__dict__["_updated_attrs"].copy()
        d.update(self.__dict__["_attrs"])
        d.update(self.__dict__["_parent_attrs"])
        return d


class RESTObjectList(object):
    """Generator object representing a list of RESTObject's.

    This generator uses the Gitlab pagination system to fetch new data when
    required.

    Note: you should not instantiate such objects, they are returned by calls
    to RESTManager.list()

    Args:
        manager: Manager to attach to the created objects
        obj_cls: Type of objects to create from the json data
        _list: A GitlabList object
    """

    def __init__(self, manager, obj_cls, _list):
        """Creates an objects list from a GitlabList.

        You should not create objects of this type, but use managers list()
        methods instead.

        Args:
            manager: the RESTManager to attach to the objects
            obj_cls: the class of the created objects
            _list: the GitlabList holding the data
        """
        self.manager = manager
        self._obj_cls = obj_cls
        self._list = _list

    def __iter__(self) -> "RESTObjectList":
        return self

    def __len__(self) -> int:
        return len(self._list)

    def __next__(self) -> RESTObject:
        return self.next()

    def next(self) -> RESTObject:
        data = self._list.next()
        return self._obj_cls(self.manager, data, created_from_list=True)

    @property
    def current_page(self) -> int:
        """The current page number."""
        return self._list.current_page

    @property
    def prev_page(self) -> Optional[int]:
        """The previous page number.

        If None, the current page is the first.
        """
        return self._list.prev_page

    @property
    def next_page(self) -> Optional[int]:
        """The next page number.

        If None, the current page is the last.
        """
        return self._list.next_page

    @property
    def per_page(self) -> int:
        """The number of items per page."""
        return self._list.per_page

    @property
    def total_pages(self) -> Optional[int]:
        """The total number of pages."""
        return self._list.total_pages

    @property
    def total(self) -> Optional[int]:
        """The total number of items."""
        return self._list.total


class RequiredOptional(NamedTuple):
    required: Tuple[str, ...] = tuple()
    optional: Tuple[str, ...] = tuple()


class RESTManager(object):
    """Base class for CRUD operations on objects.

    Derived class must define ``_path`` and ``_obj_cls``.

    ``_path``: Base URL path on which requests will be sent (e.g. '/projects')
    ``_obj_cls``: The class of objects that will be created
    """

    _create_attrs = RequiredOptional()
    _update_attrs = RequiredOptional()
    _path = None
    _obj_cls = None
    _from_parent_attrs = {}
    _types = {}

    _computed_path: Optional[str]
    _parent: Optional[RESTObject]
    _parent_attrs = {}
    csgenome: CSGenome

    def __init__(self, csg: CSGenome, parent=None) -> None:
        """REST manager constructor.
        Args:
            gl: :class:`~gitlab.Gitlab` connection to use to make requests.
            parent: REST object to which the manager is attached.
        """
        self.csgenome = csg
        self._parent = parent  # for nested managers, NOTE: might come in handy with country to systems
        # self._computed_path = self._compute_path()
        self._computed_path = None

    @property
    def get_parent_attrs(self):  # a getter since the private field
        return self._parent_attrs

    # def _compute_path(self, path=None):
    #     self._parent_attrs = {}
    #     if path is not None:
    #         path = self._path
    #     if path is None:
    #         return None
    #     if self._parent is None or not self._from_parent_attrs:
    #         return path

    #     data = {
    #         self_attr: getattr(self._parent, parent_attr, None)
    #         for self_attr, parent_attr in self._from_parent_attrs.items()
    #     }
    #     self._parent_attrs = data
    #     return path.format(**data)

    @property
    def get_path(self):  # getter for getting the computed path
        return self._computed_path



from types import ModuleType
from typing import (Any, Dict, Optional, Type, Union)

import clientAPI.base as base
from .client import CSGList, CSGenome
import clientAPI.utils as exc


__all__ = [
    "GetMixin",
    "GetWithoutIdMixin",
    "RefreshMixin",
    "ListMixin",
    "RetrieveMixin",
    "CreateMixin",
    "UpdateMixin",
    "SetMixin",
    "DeleteMixin",
    "CRUDMixin",
    "NoUpdateMixin",
    "SaveMixin",
    "ObjectDeleteMixin",
    "UserAgentDetailMixin",
    "AccessRequestMixin",
    "DownloadMixin",
    "SubscribableMixin",
    "TodoMixin",
    "TimeTrackingMixin",
    "ParticipantsMixin",
    "BadgeRenderMixin",
]


_RestManagerBase = base.RESTManager
_RestObjectBase = base.RESTObject


class GetMixin(_RestManagerBase):
    # @exc.on_http_error(exc.CSGServerError)
    def get(self, uid=None, params=None):
        """Retrieve a single object.
        Args:
            uid: uid of the object to retrieve
            params: additional query parameters (e.g. filter, column)
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The generated RESTObject.

        Raises:
            // GitlabAuthenticationError: If authentication is not correct
            CSGServerError: If the server cannot perform the request
        """
        try:
            assert (uid != None), "uid cannot be empty"
            assert (isinstance(uid, int)), "uid has to be an integer"
            assert ((not params) or ("filter" not in params.keys())), "filters are not allowed"
            ## easy name stuff here ##
            incoming_filter = params or {}
            # for k,v in kwargs:
            #     if k in self._create_attrs.required:
            #         incoming_filter[k] = v
            server_data = self.csgenome.http_get(path=f'{self._path}/{uid}', params=incoming_filter)
            # print(f'server data: {server_data}')

            try:
                if 'data' in server_data.keys():
                    res = server_data['data']  # TODO: very rigid implementation
                    return self._obj_cls(self, res)
                else:
                    # print(server_data)
                    return server_data['errors']
            except KeyError:
                return server_data
        except AssertionError as e:
            print(f"Input error: {e}")
            return None


class ListMixin(_RestManagerBase):
    _obj_cls: Optional[Type[base.RESTObject]]
    _path: Optional[str]
    csgenome: CSGenome

    def list(self, columns=None, exact=None, order_by=None, desc=False, **kwargs):  # TODO: maybe support none json output in the future
        """Retrieve a list of objects.
        Args:
            all: If True, return all the items, without pagination
            per_page: Number of items to retrieve per request
            page: ID of the page to return (starts with page 1)
            as_list: If set to False and no pagination option is
                defined, return a generator instead of a list
            **kwargs: Extra options to send to the server (e.g. sudo)
        Returns:
            The list of objects, or a generator if `as_list` is False
        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabListError: If the server cannot perform the request
        """
        query_param = self._parse_params(kwargs, columns=columns, exact=exact, order_by=order_by, desc=desc)
        for k, v in query_param.items():
            query_param[k] = v.strip()
        obj, error = self.csgenome.http_list(self._path, query_param)
        if error:
            return obj
        return [self._obj_cls(self, item, created_from_list=True) for item in obj]

    def list_first(self, dictionary=None, columns=None, exact=False, report_count=False, order_by=None, desc=False, **kwargs):
        """
        Return a single object that matches provided properties
        You can pass in a dictionary containing key value pairs to search and filter records by,
        or use keyword parameters as key value pairs instead
        Kwargs:
            exact: True/False (default False)
                If exact, then string fields are case sensitive
            report_count: True/False (default False)
               If true, then returns a tuple (obj, num) where num is the number of 
               records in table that matched the specified search parameters
         **kwargs: values for columns of the table you want to search for (if a dictionary if
         passed in, these kwargs are ignored)
        """
        query_param = self._parse_params(dictionary or kwargs, columns=columns,
                                         exact=exact, order_by=order_by, desc=desc)
        for k, v in query_param.items():
            if isinstance(v, str):
                query_param[k] = v.strip()

        obj, error = self.csgenome.http_list(self._path, query_param)
        # return self._obj_cls(self, obj[0])
        if error or not obj:
            return obj
        if report_count:
            return self._obj_cls(self, obj[0]), len(obj)
        return self._obj_cls(self, obj[0])

    def _parse_params(self, params: Dict, columns: Union[list, str], exact: bool, order_by: str, desc: bool) -> Dict:
        result = {}
        if columns:
            # add column parameter to result, which can either be in form
            # 'column': 'single_col'   OR
            # 'column': ['list', 'of', 'cols']
            result['columns'] = ','.join(columns) if isinstance(columns, list) else str(columns)
        if exact:
            result['filter_mode'] = 'exact'
        for k, v in params.items():
            k = k.replace('__', '.')  # for foreign key fields
            if v is None:
                continue
            if 'filter' in result:  # else we have a combined filter field
                value = ','.join(v) if isinstance(v, list) else v
                result['filter'] += f';{k}:{value}'
            else:  # first filter
                value = ','.join(v) if isinstance(v, list) else v
                result['filter'] = f'{k}:{value}'
        if order_by:
            result['order_by'] = order_by
            if desc:
                result['desc'] = True
        return result

    def gen(self, columns=None, exact=None, order_by=None, desc=False, **kwargs):
        query_param = self._parse_params(kwargs, columns=columns, exact=exact, order_by=order_by, desc=desc)
        for k, v in query_param.items():
            query_param[k] = v.strip()
        if "page" not in query_param.keys():
            query_param['page'] = 1
        if "limit" not in query_param.keys():
            query_param['limit'] = 50
        csg_list = CSGList(self.csgenome, self._path, query_data=query_param)

        return csg_list.gen(lambda item: self._obj_cls(self, item, created_from_list=True))


class CreateMixin(_RestManagerBase):
    _path: Optional[str]
    csgenome: CSGenome

    def _check_missing_create_attrs(self, data):
        missing = []
        for attr in self._create_attrs.required:  # passed in from obj
            if attr not in data:
                missing.append(attr)
        if missing:
            raise AttributeError(f"Missing attributes: {', '.join(missing)}")

    def create(self, data=None, return_response=False):
        """Create a new object.

        Args:
            data: parameters to send to the server to create the
                         resource
            return_response: If true, return the entire response body 
                         dictionary instead of just the data on success 
                         (default False)
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
        Tuple (resp_dict, is_error) containing:
            A dictionary containing server data on success and error
            information on failure 
            and a boolean if error occured or not


        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabCreateError: If the server cannot perform the request
        """
        data = data or {}

        self._check_missing_create_attrs(data)

        path = self._path  # NOTE: might not be necessary, consider

        data = {k: v for k, v in data.items() if v is not None}
        server_data = self.csgenome.http_post(path, post_data=data)

        if 'errors' in server_data:
            # then we have error
            return server_data['errors'], True
        elif 'status' in server_data and not str(server_data['status']).startswith('2'):
            return server_data, True
        elif return_response:
            return server_data, False
        else:
            return server_data['data'], False  # we do not return an object when added


class UpdateMixin(_RestManagerBase):
    _computed_path: Optional[str]
    _from_parent_attrs: Dict[str, Any]
    _obj_cls: Optional[Type[base.RESTObject]]
    _parent: Optional[base.RESTObject]
    _parent_attrs: Dict[str, Any]
    _path: Optional[str]
    _update_uses_post: bool = False
    csgenome: CSGenome

    def _check_missing_update_attrs(self, data):
        # Remove the id field from the required list as it was previously moved
        # to the http path.
        required = tuple(
            [k for k in self._update_attrs.required if k != self._obj_cls._id_attr]
        )
        missing = []
        for attr in required:
            if attr not in data:
                missing.append(attr)
        if missing:
            raise AttributeError(f"Missing attributes: {', '.join(missing)}")

    def update(self, uid, new_data=None):
        """Update an object on the server.

        Args:
            id: ID of the object to update (can be None if not required)
            new_data: the update data for the object
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The new object data (*not* a RESTObject)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server cannot perform the request
        """
        new_data = new_data or {}

        if uid is None:
            raise AttributeError(f"Missing uid.")
        else:
            path = f"{self._path}/{uid}"

        self._check_missing_update_attrs(new_data)

        server_data = self.csgenome.http_put(path, put_data=new_data)

        if 'data' in server_data.keys():
            res = server_data['data']  # TODO: very rigid implementation
            return self._obj_cls(self, res)
        else:
            return server_data['errors']


# class SetMixin(_RestManagerBase):
#     _computed_path: Optional[str]
#     _from_parent_attrs: Dict[str, Any]
#     _obj_cls: Optional[Type[base.RESTObject]]
#     _parent: Optional[base.RESTObject]
#     _parent_attrs: Dict[str, Any]
#     _path: Optional[str]
#     gitlab: gitlab.Gitlab

#     @exc.on_http_error(exc.GitlabSetError)
#     def set(self, key: str, value: str, **kwargs: Any) -> base.RESTObject:
#         """Create or update the object.

#         Args:
#             key: The key of the object to create/update
#             value: The value to set for the object
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabSetError: If an error occurred

#         Returns:
#             The created/updated attribute
#         """
#         path = f"{self.path}/{utils.EncodedId(key)}"
#         data = {"value": value}
#         server_data = self.gitlab.http_put(path, post_data=data, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(server_data, requests.Response)
#             assert self._obj_cls is not None
#         return self._obj_cls(self, server_data)


class DeleteMixin(_RestManagerBase):
    _path: Optional[str]
    csgenome: CSGenome

    def delete(self, uid):
        """Delete an object on the server.

        Args:
            id: ID of the object to delete
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the server cannot perform the request
        """
        if uid is None:
            raise AttributeError(f"Missing uid.")
        else:
            path = f"{self._path}/{uid}"
        server_data = self.csgenome.http_delete(path)
        return self._obj_cls(self, server_data)


# class CRUDMixin(GetMixin, ListMixin, CreateMixin, UpdateMixin, DeleteMixin):
#     _computed_path: Optional[str]
#     _from_parent_attrs: Dict[str, Any]
#     _obj_cls: Optional[Type[base.RESTObject]]
#     _parent: Optional[base.RESTObject]
#     _parent_attrs: Dict[str, Any]
#     _path: Optional[str]
#     gitlab: gitlab.Gitlab

#     pass


# class NoUpdateMixin(GetMixin, ListMixin, CreateMixin, DeleteMixin):
#     _computed_path: Optional[str]
#     _from_parent_attrs: Dict[str, Any]
#     _obj_cls: Optional[Type[base.RESTObject]]
#     _parent: Optional[base.RESTObject]
#     _parent_attrs: Dict[str, Any]
#     _path: Optional[str]
#     gitlab: gitlab.Gitlab

#     pass


class SaveMixin(_RestObjectBase):
    """Mixin for RESTObject's that can be updated."""

    _id_attr: Optional[str]
    _attrs: Dict[str, Any]
    _module: ModuleType
    _parent_attrs: Dict[str, Any]
    _updated_attrs: Dict[str, Any]
    manager: base.RESTManager

    def _get_updated_data(self):
        updated_data = {}
        for attr in self.manager._update_attrs.required:
            # Get everything required, no matter if it's been updated
            updated_data[attr] = getattr(self, attr)
        # Add the updated attributes
        updated_data.update(self._updated_attrs)

        return updated_data

    def save(self):
        """Save the changes made to the object to the server.

        The object is updated to match what the server returns.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The new object data (*not* a RESTObject)

        Raise:
            GitlabAuthenticationError: If authentication is not correct
            GitlabUpdateError: If the server cannot perform the request
        """
        updated_data = self._get_updated_data()
        # Nothing to update. Server fails if sent an empty dict.
        if not updated_data:
            return None

        # call the manager
        obj_id = self.encoded_id
        server_data = self.manager.update(uid=obj_id, new_data=updated_data)
        self._update_attrs(server_data)
        return server_data


class ObjectDeleteMixin(_RestObjectBase):
    """Mixin for RESTObject's that can be deleted."""

    _id_attr: Optional[str]
    _attrs: Dict[str, Any]
    _module: ModuleType
    _parent_attrs: Dict[str, Any]
    _updated_attrs: Dict[str, Any]
    manager: base.RESTManager

    def delete(self, **kwargs: Any) -> None:
        """Delete the object from the server.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabDeleteError: If the server cannot perform the request
        """
        self.manager.delete(self.encoded_id, **kwargs)


# class UserAgentDetailMixin(_RestObjectBase):
#     _id_attr: Optional[str]
#     _attrs: Dict[str, Any]
#     _module: ModuleType
#     _parent_attrs: Dict[str, Any]
#     _updated_attrs: Dict[str, Any]
#     manager: base.RESTManager

#     @cli.register_custom_action(("Snippet", "ProjectSnippet", "ProjectIssue"))
#     @exc.on_http_error(exc.GitlabGetError)
#     def user_agent_detail(self, **kwargs: Any) -> Dict[str, Any]:
#         """Get the user agent detail.

#         Args:
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabGetError: If the server cannot perform the request
#         """
#         path = f"{self.manager.path}/{self.encoded_id}/user_agent_detail"
#         result = self.manager.gitlab.http_get(path, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result


# class AccessRequestMixin(_RestObjectBase):
#     _id_attr: Optional[str]
#     _attrs: Dict[str, Any]
#     _module: ModuleType
#     _parent_attrs: Dict[str, Any]
#     _updated_attrs: Dict[str, Any]
#     manager: base.RESTManager

#     @cli.register_custom_action(
#         ("ProjectAccessRequest", "GroupAccessRequest"), tuple(), ("access_level",)
#     )
#     @exc.on_http_error(exc.GitlabUpdateError)
#     def approve(
#         self, access_level: int = gitlab.const.DEVELOPER_ACCESS, **kwargs: Any
#     ) -> None:
#         """Approve an access request.

#         Args:
#             access_level: The access level for the user
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabUpdateError: If the server fails to perform the request
#         """

#         path = f"{self.manager.path}/{self.encoded_id}/approve"
#         data = {"access_level": access_level}
#         server_data = self.manager.gitlab.http_put(path, post_data=data, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(server_data, requests.Response)
#         self._update_attrs(server_data)


# class DownloadMixin(_RestObjectBase):
#     _id_attr: Optional[str]
#     _attrs: Dict[str, Any]
#     _module: ModuleType
#     _parent_attrs: Dict[str, Any]
#     _updated_attrs: Dict[str, Any]
#     manager: base.RESTManager

#     @cli.register_custom_action(("GroupExport", "ProjectExport"))
#     @exc.on_http_error(exc.GitlabGetError)
#     def download(
#         self,
#         streamed: bool = False,
#         action: Optional[Callable] = None,
#         chunk_size: int = 1024,
#         **kwargs: Any,
#     ) -> Optional[bytes]:
#         """Download the archive of a resource export.

#         Args:
#             streamed: If True the data will be processed by chunks of
#                 `chunk_size` and each chunk is passed to `action` for
#                 treatment
#             action: Callable responsible of dealing with chunk of
#                 data
#             chunk_size: Size of each chunk
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabGetError: If the server failed to perform the request

#         Returns:
#             The blob content if streamed is False, None otherwise
#         """
#         path = f"{self.manager.path}/download"
#         result = self.manager.gitlab.http_get(
#             path, streamed=streamed, raw=True, **kwargs
#         )
#         if TYPE_CHECKING:
#             assert isinstance(result, requests.Response)
#         return utils.response_content(result, streamed, action, chunk_size)


# class SubscribableMixin(_RestObjectBase):
#     _id_attr: Optional[str]
#     _attrs: Dict[str, Any]
#     _module: ModuleType
#     _parent_attrs: Dict[str, Any]
#     _updated_attrs: Dict[str, Any]
#     manager: base.RESTManager

#     @cli.register_custom_action(
#         ("ProjectIssue", "ProjectMergeRequest", "ProjectLabel", "GroupLabel")
#     )
#     @exc.on_http_error(exc.GitlabSubscribeError)
#     def subscribe(self, **kwargs: Any) -> None:
#         """Subscribe to the object notifications.

#         Args:
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabSubscribeError: If the subscription cannot be done
#         """
#         path = f"{self.manager.path}/{self.encoded_id}/subscribe"
#         server_data = self.manager.gitlab.http_post(path, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(server_data, requests.Response)
#         self._update_attrs(server_data)

#     @cli.register_custom_action(
#         ("ProjectIssue", "ProjectMergeRequest", "ProjectLabel", "GroupLabel")
#     )
#     @exc.on_http_error(exc.GitlabUnsubscribeError)
#     def unsubscribe(self, **kwargs: Any) -> None:
#         """Unsubscribe from the object notifications.

#         Args:
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabUnsubscribeError: If the unsubscription cannot be done
#         """
#         path = f"{self.manager.path}/{self.encoded_id}/unsubscribe"
#         server_data = self.manager.gitlab.http_post(path, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(server_data, requests.Response)
#         self._update_attrs(server_data)


# class TodoMixin(_RestObjectBase):
#     _id_attr: Optional[str]
#     _attrs: Dict[str, Any]
#     _module: ModuleType
#     _parent_attrs: Dict[str, Any]
#     _updated_attrs: Dict[str, Any]
#     manager: base.RESTManager

#     @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"))
#     @exc.on_http_error(exc.GitlabTodoError)
#     def todo(self, **kwargs: Any) -> None:
#         """Create a todo associated to the object.

#         Args:
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabTodoError: If the todo cannot be set
#         """
#         path = f"{self.manager.path}/{self.encoded_id}/todo"
#         self.manager.gitlab.http_post(path, **kwargs)


# class TimeTrackingMixin(_RestObjectBase):
#     _id_attr: Optional[str]
#     _attrs: Dict[str, Any]
#     _module: ModuleType
#     _parent_attrs: Dict[str, Any]
#     _updated_attrs: Dict[str, Any]
#     manager: base.RESTManager

#     @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"))
#     @exc.on_http_error(exc.GitlabTimeTrackingError)
#     def time_stats(self, **kwargs: Any) -> Dict[str, Any]:
#         """Get time stats for the object.

#         Args:
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabTimeTrackingError: If the time tracking update cannot be done
#         """
#         # Use the existing time_stats attribute if it exist, otherwise make an
#         # API call
#         if "time_stats" in self.attributes:
#             return self.attributes["time_stats"]

#         path = f"{self.manager.path}/{self.encoded_id}/time_stats"
#         result = self.manager.gitlab.http_get(path, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result

#     @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"), ("duration",))
#     @exc.on_http_error(exc.GitlabTimeTrackingError)
#     def time_estimate(self, duration: str, **kwargs: Any) -> Dict[str, Any]:
#         """Set an estimated time of work for the object.

#         Args:
#             duration: Duration in human format (e.g. 3h30)
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabTimeTrackingError: If the time tracking update cannot be done
#         """
#         path = f"{self.manager.path}/{self.encoded_id}/time_estimate"
#         data = {"duration": duration}
#         result = self.manager.gitlab.http_post(path, post_data=data, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result

#     @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"))
#     @exc.on_http_error(exc.GitlabTimeTrackingError)
#     def reset_time_estimate(self, **kwargs: Any) -> Dict[str, Any]:
#         """Resets estimated time for the object to 0 seconds.

#         Args:
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabTimeTrackingError: If the time tracking update cannot be done
#         """
#         path = f"{self.manager.path}/{self.encoded_id}/reset_time_estimate"
#         result = self.manager.gitlab.http_post(path, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result

#     @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"), ("duration",))
#     @exc.on_http_error(exc.GitlabTimeTrackingError)
#     def add_spent_time(self, duration: str, **kwargs: Any) -> Dict[str, Any]:
#         """Add time spent working on the object.

#         Args:
#             duration: Duration in human format (e.g. 3h30)
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabTimeTrackingError: If the time tracking update cannot be done
#         """
#         path = f"{self.manager.path}/{self.encoded_id}/add_spent_time"
#         data = {"duration": duration}
#         result = self.manager.gitlab.http_post(path, post_data=data, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result

#     @cli.register_custom_action(("ProjectIssue", "ProjectMergeRequest"))
#     @exc.on_http_error(exc.GitlabTimeTrackingError)
#     def reset_spent_time(self, **kwargs: Any) -> Dict[str, Any]:
#         """Resets the time spent working on the object.

#         Args:
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabTimeTrackingError: If the time tracking update cannot be done
#         """
#         path = f"{self.manager.path}/{self.encoded_id}/reset_spent_time"
#         result = self.manager.gitlab.http_post(path, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result


# class ParticipantsMixin(_RestObjectBase):
#     _id_attr: Optional[str]
#     _attrs: Dict[str, Any]
#     _module: ModuleType
#     _parent_attrs: Dict[str, Any]
#     _updated_attrs: Dict[str, Any]
#     manager: base.RESTManager

#     @cli.register_custom_action(("ProjectMergeRequest", "ProjectIssue"))
#     @exc.on_http_error(exc.GitlabListError)
#     def participants(self, **kwargs: Any) -> Dict[str, Any]:
#         """List the participants.

#         Args:
#             all: If True, return all the items, without pagination
#             per_page: Number of items to retrieve per request
#             page: ID of the page to return (starts with page 1)
#             as_list: If set to False and no pagination option is
#                 defined, return a generator instead of a list
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabListError: If the list could not be retrieved

#         Returns:
#             The list of participants
#         """

#         path = f"{self.manager.path}/{self.encoded_id}/participants"
#         result = self.manager.gitlab.http_get(path, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result


# class BadgeRenderMixin(_RestManagerBase):
#     @cli.register_custom_action(
#         ("GroupBadgeManager", "ProjectBadgeManager"), ("link_url", "image_url")
#     )
#     @exc.on_http_error(exc.GitlabRenderError)
#     def render(self, link_url: str, image_url: str, **kwargs: Any) -> Dict[str, Any]:
#         """Preview link_url and image_url after interpolation.

#         Args:
#             link_url: URL of the badge link
#             image_url: URL of the badge image
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabRenderError: If the rendering failed

#         Returns:
#             The rendering properties
#         """
#         path = f"{self.path}/render"
#         data = {"link_url": link_url, "image_url": image_url}
#         result = self.gitlab.http_get(path, data, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result


# class PromoteMixin(_RestObjectBase):
#     _id_attr: Optional[str]
#     _attrs: Dict[str, Any]
#     _module: ModuleType
#     _parent_attrs: Dict[str, Any]
#     _updated_attrs: Dict[str, Any]
#     _update_uses_post: bool = False
#     manager: base.RESTManager

#     def _get_update_method(
#         self,
#     ) -> Callable[..., Union[Dict[str, Any], requests.Response]]:
#         """Return the HTTP method to use.

#         Returns:
#             http_put (default) or http_post
#         """
#         if self._update_uses_post:
#             http_method = self.manager.gitlab.http_post
#         else:
#             http_method = self.manager.gitlab.http_put
#         return http_method

#     @exc.on_http_error(exc.GitlabPromoteError)
#     def promote(self, **kwargs: Any) -> Dict[str, Any]:
#         """Promote the item.

#         Args:
#             **kwargs: Extra options to send to the server (e.g. sudo)

#         Raises:
#             GitlabAuthenticationError: If authentication is not correct
#             GitlabPromoteError: If the item could not be promoted
#             GitlabParsingError: If the json data could not be parsed

#         Returns:
#             The updated object data (*not* a RESTObject)
#         """

#         path = f"{self.manager.path}/{self.encoded_id}/promote"
#         http_method = self._get_update_method()
#         result = http_method(path, **kwargs)
#         if TYPE_CHECKING:
#             assert not isinstance(result, requests.Response)
#         return result



from typing import cast
from ..base import RESTManager, RESTObject, RequiredOptional
from ..mixin import GetMixin, CreateMixin, UpdateMixin, SaveMixin, ListMixin
from ..utils import handle_client_exception


__all__ = [
    "Pointer",
    "PointerManager",
]


class Pointer(SaveMixin, RESTObject):
    _id_attr = "uid"
    # pass


# NOTE: even tho Delete Mixin is integrated, since pointer is a look up table,
# our API does not support DELETE for it
class PointerManager(GetMixin, CreateMixin, UpdateMixin, ListMixin, RESTManager):
    _path = "benchmarks/pointer"
    _obj_cls = Pointer
    _create_attrs = RequiredOptional(required=("name",)) 
    #TODO: figure out if the auth section should be passed in here, 

    def get(self, uid=None, params=None):
        # NOTE: the casting allows each object to follow the logic implemented in the
        # MIXIN utils w/o the need to repeating it each time. it also allows customization
        # in each obj
        return cast(Pointer, super(PointerManager, self).get(uid=uid, params=params))





from typing import cast
from ..base import RESTManager, RESTObject, RequiredOptional
from ..mixin import GetMixin, CreateMixin, UpdateMixin, SaveMixin, ListMixin, ObjectDeleteMixin, DeleteMixin
from ..utils import handle_client_exception

__all__ = [
    "Systems",
    "SystemsManager",
]


class Systems(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "uid"
    # pass


class SystemsManager(GetMixin, CreateMixin, UpdateMixin, ListMixin, DeleteMixin, RESTManager):
    _path = "systems"
    _obj_cls = Systems
    _create_attrs = RequiredOptional(required=(), optional=("moniker", "system_model",\
        "system_family_id", "architecture_id", "system_manufacturer_id", "computer", \
            "site", "application_id", "country_id", "region_id", "continent_id", "segment_id", \
            "system_address", "latitude", "longitude"))
    #TODO: figure out if the auth section should be passed in here,
    # if so we need the func in mixin to refactor attrs before calling POST
    _update_attrs = RequiredOptional(required=(), optional=("moniker", "system_model",\
        "system_family_id", "architecture_id", "system_manufacturer_id", "computer", \
            "site", "application_id", "country_id", "region_id", "continent_id", "segment_id", \
            "system_address", "latitude", "longitude"))

    def get(self, uid=None, params=None):
        # NOTE: the casting allows each object to follow the logic implemented in the
        # MIXIN utils w/o the need to repeating it each time. it also allows customization
        # in each obj
        return cast(Systems, super(SystemsManager, self).get(uid=uid, params=params))


from typing import cast
from ..base import RESTManager, RESTObject, RequiredOptional
from ..mixin import DeleteMixin, GetMixin, CreateMixin, ObjectDeleteMixin, UpdateMixin, SaveMixin, ListMixin


__all__ = [
    "Spec2017",
    "Spec2017Manager",
]


class Spec2017(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "uid"
    # pass


class Spec2017Manager(GetMixin, CreateMixin, UpdateMixin, ListMixin, DeleteMixin, RESTManager):
    _path = "benchmarks/spec2017"
    _obj_cls = Spec2017
    _create_attrs = RequiredOptional(required=("information_source", 'benchmark_type'))
    # TODO: figure out if the auth section should be passed in here,
    # if so we need the func in mixin to refactor attrs before calling POST

    def get(self, uid=None, params=None):
        # NOTE: the casting allows each object to follow the logic implemented in the
        # MIXIN utils w/o the need to repeating it each time. it also allows customization
        # in each obj
        return cast(Spec2017, super(Spec2017Manager, self).get(uid=uid, params=params))



from typing import cast
from ..base import RESTManager, RESTObject, RequiredOptional
from ..mixin import DeleteMixin, GetMixin, CreateMixin, UpdateMixin, SaveMixin, ListMixin, ObjectDeleteMixin
from ..utils import handle_client_exception

__all__ = [
    "Green500",
    "Green500Manager",
]


class Green500(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "uid"
    # pass


class Green500Manager(GetMixin, CreateMixin, DeleteMixin, UpdateMixin, ListMixin, RESTManager):
    _path = "benchmarks/green500"
    _obj_cls = Green500
    _create_attrs = RequiredOptional(required=("information_source",), optional=("benchmark_type", "green500_rank", "green500_previous_rank", \
        "top500_rank", "install_year", "publication_date", "power", "power_quality_level", \
            "power_source", "measured_cores", "total_cores", "optimized_run_hpl", \
                "optimized_run_peak_power", "submission_type_id", "mflops_per_watt", \
                    "efficiency", "r_max", "r_peak", "n_max", "note"))
    # TODO: figure out if the auth section should be passed in here,
    # if so we need the func in mixin to refactor attrs before calling POST
    _update_attrs = RequiredOptional(required=(), optional=("benchmark_type", "green500_rank", "green500_previous_rank", \
        "top500_rank", "information_source", "install_year", "publication_date", "power", "power_quality_level", \
            "power_source", "measured_cores", "total_cores", "optimized_run_hpl", \
                "optimized_run_peak_power", "submission_type_id", "mflops_per_watt", \
                    "efficiency", "r_max", "r_peak", "n_max", "note"))

    def get(self, uid=None, params=None):
        # NOTE: the casting allows each object to follow the logic implemented in the
        # MIXIN utils w/o the need to repeating it each time. it also allows customization
        # in each obj
        return cast(Green500, super(Green500Manager, self).get(uid=uid, params=params))




# TODO: in this file we are going to follow the client.py to create obejcts to represent the
# csg server connection. cli.py is an addition for custom features for gitlab, optional for us

# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Wrapper for the GitLab API."""
import json

from .utils import handle_client_exception
from contextlib import nullcontext
from typing import Any, Optional

import requests
import requests.utils
import sys
print(sys.path)


REDIRECT_MSG = (
    "python-gitlab detected a {status_code} ({reason!r}) redirection. You must update "
    "your GitLab URL to the correct URL to avoid issues. The redirection was from: "
    "{source!r} to {target!r}"
)

DEFAULT_BASE_UTL = "http://127.0.0.1:5000/"  # can change in the future
CONTENT_TYPE = "application/json"


class CSGenome(object):
    """Represents a GitLab server connection.

    Args:
        url: The URL of the GitLab server (defaults to https://gitlab.com).
        private_token: The user private token
        oauth_token: An oauth token
        job_token: A CI job token
        ssl_verify: Whether SSL certificates should be validated. If
            the value is a string, it is the path to a CA file used for
            certificate validation.
        timeout: Timeout to use for requests to the GitLab server.
        http_username: Username for HTTP authentication
        http_password: Password for HTTP authentication
        api_version: Gitlab API version to use (support for 4 only)
        pagination: Can be set to 'keyset' to use keyset pagination
        order_by: Set order_by globally
        user_agent: A custom user agent to use for making HTTP requests.
        retry_transient_errors: Whether to retry after 500, 502, 503, or
            504 responses. Defaults to False.
    """

    def __init__(self, url=None, auth_token=None, email=None, password=None, new_password=None, per_page=50):
        # http_username: Optional[str] = None, # NOTE: depends on the future of auth
        # http_password: Optional[str] = None, # NOTE: same as above
        # api_version: str = "4", # NOTE: not in the near future lol
        # pagination: Optional[str] = None, # NOTE: with the current api implementation, we can only support num per page. this would be good to have tho
        # order_by: Optional[str] = None, # NOTE: also nice to have, don't know how useful it will be
        # -------------- params ---------------

        # self._url = f"{self._base_url}/api/v{api_version}" # NOTE: I don't think we have these, or will have them
        self._url = self._get_base_url(url)
        #: Timeout to use for requests to gitlab server
        # self.timeout = timeout # NOTE: might introduce in the future
        self.headers = {'email': email, 'password': password, "Content-type": CONTENT_TYPE}
        self.data = {'email': email, 'password': password, 'new_password': new_password}
        self.email = email
        self.password = password
        self.new_password = new_password
        self.cookies = {'token': auth_token}
        self.auth_token = auth_token  # NOTE: not sure if it will be used in the future
        self.session = requests.Session()
        self.per_page = per_page
        self._set_auth_info()

        # ------------ Objects ------------
        # NOTE: We must delay import of gitlab.v4.objects until now or
        # otherwise it will cause circular import errors
        # import gitlab.v4.objects
        # import .objects.country as country_object
        # import .objects.codename as codename_object
        from .objects import codename, country, file_system, \
            firmware, floating_point_unit, interconnect_family, manufacturer, microarchitectures, \
            operating_system, other_software, pointer, system_state
        from .objects import country, genera, green500, isas, ontology_type, \
            processor_hierarchy, processors, spec1995, spec2000, spec2006, spec2017, \
            top500, top500_ranks, processor_configuration, has_processor, memory_configuration, \
            has_memory, accelerator_configuration, has_accelerator, interconnect_configuration, \
            has_interconnect, operating_system_configuration, has_operating_system, system_configuration
        from .objects import country, application, architecture,\
            continent, region, segment, system_family, system_manufacturer, systems, interconnect, \
            benchmarks, used_to_run

        self.used_to_run = used_to_run.UsedToRunManager(self)
        self.benchmarks = benchmarks.BenchmarksManager(self)
        self.codename = codename.CodenameManager(self)
        self.country = country.CountryManager(self)
        self.file_system = file_system.FileSystemManager(self)
        self.firmware = firmware.FirmwareManager(self)
        self.floating_point_unit = floating_point_unit.FloatingPointUnitManager(self)
        self.interconnect = interconnect.InterconnectManager(self)
        self.interconnect_family = interconnect_family.InterconnectFamilyManager(self)
        self.manufacturer = manufacturer.ManufacturerManager(self)
        self.microarchitectures = microarchitectures.MicroarchitecturesManager(self)
        self.operating_system = operating_system.OperatingSystemManager(self)
        self.other_software = other_software.OtherSoftwareManager(self)
        self.pointer = pointer.PointerManager(self)
        self.system_state = system_state.SystemStateManager(self)
        self.genera = genera.GeneraManager(self)
        self.green500 = green500.Green500Manager(self)
        self.isas = isas.IsasManager(self)
        self.ontology_type = ontology_type.OntologyTypeManager(self)
        self.processor_hierarchy = processor_hierarchy.ProcessorHierarchyManager(self)
        self.processors = processors.ProcessorsManager(self)
        self.processor_configuration = processor_configuration.ProcessorConfigurationManager(self)
        self.has_processor = has_processor.HasProcessorManager(self)
        self.memory_configuration = memory_configuration.MemoryConfigurationManager(self)
        self.has_memory = has_memory.HasMemoryManager(self)
        self.accelerator_configuration = accelerator_configuration.AcceleratorConfigurationManager(self)
        self.has_accelerator = has_accelerator.HasAcceleratorManager(self)
        self.interconnect_configuration = interconnect_configuration.InterconnectConfigurationManager(self)
        self.has_interconnect = has_interconnect.HasInterconnectManager(self)
        self.operating_system_configuration = operating_system_configuration.OperatingSystemConfigurationManager(self)
        self.has_operating_system = has_operating_system.HasOperatingSystemManager(self)
        self.system_configuration = system_configuration.SystemConfigurationManager(self)
        self.spec1995 = spec1995.Spec1995Manager(self)
        self.spec2000 = spec2000.Spec2000Manager(self)
        self.spec2006 = spec2006.Spec2006Manager(self)
        self.spec2017 = spec2017.Spec2017Manager(self)
        self.top500 = top500.Top500Manager(self)
        self.top500_ranks = top500_ranks.Top500RanksManager(self)

        #self._objects = objects

        self.country = country.CountryManager(self)
        self.application = application.ApplicationManager(self)
        self.architecture = architecture.ArchitectureManager(self)
        self.continent = continent.ContinentManager(self)
        self.region = region.RegionManager(self)
        self.segment = segment.SegmentManager(self)
        self.system_family = system_family.SystemFamilyManager(self)
        self.system_manufacturer = system_manufacturer.SystemManufacturerManager(self)
        self.systems = systems.SystemsManager(self)

    def __enter__(self):
        return self

    def __exit__(self, *args: Any):
        self.session.close()

    @property
    def api_url(self) -> str:
        # NOTE: we don't have one that needs to be computed
        """The computed API base URL."""
        return self._url

    # TODO: adapt this to auth check for ppl using the python client
    def auth(self) -> None:
        """Performs an authentication using private token.

        The `user` attribute will hold a `gitlab.objects.CurrentUser` object on
        success.
        """
        # self.user = self._objects.CurrentUserManager(self).get()
        pass

    # NOTE: very slick, should reuse if we have more than one in the future

    def _set_auth_info(self):
        if (self.email and not self.password) or \
                (not self.email and self.password):
            raise ValueError("Both or none of email and password should be present")

        if not self.new_password:  # not changing password, should be header
            self.session.headers.update(self.headers)

    def enable_debug(self) -> None:
        import logging
        from http.client import HTTPConnection  # noqa

        HTTPConnection.debuglevel = 1
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

    def get_per_page(self):
        return self.per_page

    # def _get_session_opts(self) -> Dict[str, Any]:
    #     return {
    #         "headers": self.headers.copy(),
    #         "auth": self._http_auth,
    #         # "timeout": self.timeout,
    #         # "verify": self.ssl_verify,
    #         #TODO: add the other ones
    #     }

    def _get_base_url(self, url: Optional[str] = None) -> str:
        """Return the base URL with the trailing slash stripped.
        If the URL is a Falsy value, return the default URL.
        Returns:
            The base URL
        """
        if not url:
            return DEFAULT_BASE_UTL

        return url.rstrip("/")

    def _build_url(self, path: str) -> str:
        """Returns the full url from path.

        If path is already a url, return it unchanged. If it's a path, append
        it to the stored url.

        Returns:
            The full URL
        """
        if path.startswith("http://") or path.startswith("https://"):
            return path
        else:
            return f"{self._url}{path}"


# ------------------ REST functions -----------------------

    def http_request(self, verb, path, post_data=None, params=None):
        """Make an HTTP request to the Gitlab server.

        Args:
            verb: The HTTP method to call ('get', 'post', 'put', 'delete')
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            post_data: Data to send in the body (will be converted to
                              json by default)
            raw: If True, do not convert post_data to json
            streamed: Whether the data should be streamed
            files: The files to send to the server
            timeout: The timeout, in seconds, for the request
            obey_rate_limit: Whether to obey 429 Too Many Request
                                    responses. Defaults to True.
            max_retries: Max retries after 429 or transient errors,
                               set to -1 to retry forever. Defaults to 10.
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            A requests result object.

        Raises:
            GitlabHttpError: When the return code is not 2xx
        """
        url = self._build_url(path)
        params = params or {}

        result = self.session.request(
            headers={'Authorization': f'Bearer {self.auth_token}'} if self.auth_token else None,
            method=verb,
            url=url,
            json=post_data,  # specifically used for POST
            # data=data, #NOTE: so far not sure about the diff between data and json
            params=params,  # takes a json
        )
        if verb == 'post':  # NOTE: handles location header, can remove if necessary
            added_location = result.headers.get("Location") or None
            return (added_location, result)
        return result

    def http_get(self, path, params=None):
        """Make a GET request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            streamed: Whether the data should be streamed
            raw: If True do not try to parse the output as json
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            A requests result object is streamed is True or the content type is
            not json.
            The parsed json data otherwise.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        try:
            result = self.http_request("get", path, params=params)
            # print(result)
            return result.json()
        except Exception:
            return handle_client_exception(400, {'message': 'Unable to parse result.'})

    def http_list(self, path, params=None):
        """Make a GET request to the Gitlab server for list-oriented queries.
        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projects')
            query_data: Data to send as query parameters
            **kwargs: Extra options to send to the server (e.g. sudo, page,
                      per_page)
        Returns:
            A list of the objects returned by the server. If `as_list` is
            False and no pagination-related arguments (`page`, `per_page`,
            `all`) are defined then a GitlabList object (generator) is returned
            instead. This object will make API calls when needed to fetch the
            next items from the server.
        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """

        if "page" not in params.keys():
            params['page'] = 1
        if "limit" not in params.keys():
            params['limit'] = 50

        res = CSGList(self, path, query_data=params).json()
        return (res['errors'], True) if 'errors' in res.keys() else (list(res['data']), False)

    def http_post(self, path, post_data=None):
        """Make a POST request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            post_data: Data to send in the body (will be converted to
                              json by default)
            raw: If True, do not convert post_data to json
            files: The files to send to the server
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The parsed json returned by the server if json is return, else the
            raw content

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        # NOTE: we assume no query param
        post_data = post_data or {}

        try:
            location, result = self.http_request("post", path, post_data=post_data)
            if location:
                res = ({"Location": location, "data": result.json()})
                return res
            return result.json()
        except Exception:
            return handle_client_exception(400, {'message': 'Unable to parse result.'})

    def http_put(self, path, put_data=None):
        """Make a PUT request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data: Data to send as query parameters
            post_data: Data to send in the body (will be converted to
                              json by default)
            raw: If True, do not convert post_data to json
            files: The files to send to the server
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The parsed json returned by the server.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        put_data = put_data or {}

        result = self.http_request("put", path, post_data=put_data)
        try:
            return result.json()
        except Exception:
            return handle_client_exception(400, {'message': 'Unable to parse result.'})

    def http_delete(self, path):
        """Make a DELETE request to the Gitlab server.

        Args:
            path: Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The requests object.

        Raises:
            GitlabHttpError: When the return code is not 2xx
        """
        return self.http_request("delete", path)

    # @gitlab.exceptions.on_http_error(gitlab.exceptions.GitlabSearchError)
    # def search(
    #     self, scope: str, search: str, **kwargs: Any
    # ) -> Union["GitlabList", List[Dict[str, Any]]]:
    #     """Search GitLab resources matching the provided string.'

    #     Args:
    #         scope: Scope of the search
    #         search: Search string
    #         **kwargs: Extra options to send to the server (e.g. sudo)

    #     Raises:
    #         GitlabAuthenticationError: If authentication is not correct
    #         GitlabSearchError: If the server failed to perform the request

    #     Returns:
    #         A list of dicts describing the resources found.
    #     """
    #     data = {"scope": scope, "search": search}
    #     return self.http_list("/search", query_data=data, **kwargs)


class CSGList(object):
    """Generator representing a list of remote objects.

    The object handles the links returned by a query to the API, and will call
    the API again when needed.
    """
    # NOTE: we do not calculate page number, but only provide urls for pagination

    def __init__(self, csg: CSGenome, url: str, query_data: dict):
        self._csgenome = csg
        self._base_url = url
        self._original_query = query_data
        self._query(self._base_url, query_data)

        # Remove query_parameters from kwargs, which are saved via the `next` URL
        # self._kwargs.pop("query_parameters", None)

    def _query(self, url, params=None):
        try:
            result = self._csgenome.http_request("get", url, params=params)
            self._data = result.json()
        except Exception:
            self._data = {'errors': 'Could not connect to server. Check your connection or try again later'}
            return handle_client_exception(400, {'message': 'Unable to parse result.'})

        try:
            links = self._data['links']
            self._next_url = links["next"]
            self._fist_url = links["first"]
            self._last_url = links["last"]
            self._prev_url = links["prev"]
        except KeyError:
            pass  # TODO: return some error

    def __iter__(self):
        return self

    # def __next__(self):
    #     return self.next_page()

    def json(self):
        data = self._data.copy()
        return data

    def gen(self, object_creator):
        # default the same query data with next page
        for item in self._data['data']:
            yield object_creator(item)

        while self._next_url:
            self._original_query['page'] += 1
            new_query = self._original_query.copy()
            self._query(self._base_url, new_query)

            for item in self._data['data']:
                yield object_creator(item)



from typing import cast
from ..base import RESTManager, RESTObject, RequiredOptional
from ..mixin import DeleteMixin, GetMixin, CreateMixin, ListMixin, UpdateMixin, SaveMixin, ObjectDeleteMixin
from ..utils import handle_client_exception

__all__ = [
    "Isas",
    "IsasManager",
]


class Isas(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "uid"
    # pass


class IsasManager(GetMixin, CreateMixin, DeleteMixin, UpdateMixin, ListMixin, RESTManager):
    _path = "isas"
    _obj_cls = Isas
    _create_attrs = RequiredOptional(required=("name",), optional=("year", "note", "Designer_id"))
    # TODO: figure out if the auth section should be passed in here,
    # if so we need the func in mixin to refactor attrs before calling POST
    _update_attrs = RequiredOptional(required=("name",), optional=("year", "note", "Designer_id"))

    def get(self, uid=None, params=None):
        # NOTE: the casting allows each object to follow the logic implemented in the
        # MIXIN utils w/o the need to repeating it each time. it also allows customization
        # in each obj
        return cast(Isas, super(IsasManager, self).get(uid=uid, params=params))



from typing import cast
from ..base import RESTManager, RESTObject, RequiredOptional
from ..mixin import DeleteMixin, GetMixin, CreateMixin, ListMixin, UpdateMixin, SaveMixin, ObjectDeleteMixin
from ..utils import handle_client_exception

__all__ = [
    "OntologyType",
    "OntologyTypeManager",
]


class OntologyType(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "uid"
    # pass


class OntologyTypeManager(GetMixin, CreateMixin, DeleteMixin, UpdateMixin, ListMixin, RESTManager):
    _path = "ontology_type"
    _obj_cls = OntologyType
    _create_attrs = RequiredOptional(required=(), optional=("memory_type", "other_type", "generation", "standard", "module", "standard_type", "clock_rate", "cycle_time", "transfer_rate", "bandwidth", "cl_trcd_trp", "cas_latency", "voltage", "dimm_pins", "sodimm_pins", "microdimm_pins"))
    # TODO: figure out if the auth section should be passed in here,
    # if so we need the func in mixin to refactor attrs before calling POST
    _update_attrs = RequiredOptional(required=(), optional=("memory_type", "other_type", "generation", "standard", "module", "standard_type", "clock_rate", "cycle_time", "transfer_rate", "bandwidth", "cl_trcd_trp", "cas_latency", "voltage", "dimm_pins", "sodimm_pins", "microdimm_pins"))

    def get(self, uid=None, params=None):
        # NOTE: the casting allows each object to follow the logic implemented in the
        # MIXIN utils w/o the need to repeating it each time. it also allows customization
        # in each obj
        return cast(OntologyType, super(OntologyTypeManager, self).get(uid=uid, params=params))



from typing import cast
from ..base import RESTManager, RESTObject, RequiredOptional
from ..mixin import GetMixin, CreateMixin, UpdateMixin, SaveMixin, ListMixin
from ..utils import handle_client_exception


__all__ = [
    "Interconnect",
    "InterconnectManager",
]


class Interconnect(SaveMixin, RESTObject):
    _id_attr = "uid"
    # pass


# NOTE: even tho Delete Mixin is integrated, since interconnect_ is a look up table,
# our API does not support DELETE for it
class InterconnectManager(GetMixin, CreateMixin, UpdateMixin, ListMixin, RESTManager):
    _path = "interconnect"
    _obj_cls = Interconnect
    _create_attrs = RequiredOptional(required=("name",))
    # TODO: figure out if the auth section should be passed in here,
    # if so we need the func in mixin to refactor attrs before calling POST

    def get(self, uid=None, params=None):
        # NOTE: the casting allows each object to follow the logic implemented in the
        # MIXIN utils w/o the need to repeating it each time. it also allows customization
        # in each obj
        return cast(Interconnect, super(InterconnectManager, self).get(uid=uid, params=params))



from typing import cast
from ..base import RESTManager, RESTObject, RequiredOptional
from ..mixin import GetMixin, CreateMixin, UpdateMixin, SaveMixin, ListMixin

__all__ = [
    "OperatingSystemConfiguration",
    "OperatingSystemConfigurationManager",
]


class OperatingSystemConfiguration(SaveMixin, RESTObject):
    _id_attr = "uid"
    # pass


# NOTE: even tho Delete Mixin is integrated, since operating_system_configuration is a look up table,
# our API does not support DELETE for it
class OperatingSystemConfigurationManager(GetMixin, CreateMixin, UpdateMixin, ListMixin, RESTManager):
    _path = "operating_system_configurations"
    _obj_cls = OperatingSystemConfiguration
    # TODO: figure out if the auth section should be passed in here,

    def get(self, uid=None, params=None):
        # NOTE: the casting allows each object to follow the logic implemented in the
        # MIXIN utils w/o the need to repeating it each time. it also allows customization
        # in each obj
        return cast(OperatingSystemConfiguration, super(OperatingSystemConfigurationManager, self).get(uid=uid, params=params))



from flask import current_app
import traceback

def handle_client_exception(status=500, error_data={}, ex=None):
    """Exception handler for endpoints. This wraps error message
    in a common format and allows the user-specified return code.
    Use this in except block to return relvent error message to the
    client. For example:
    ```
    try:
        ...
    except orm_exception.NoResultFound:
        handle_exception(404, {'message': 'Object with uid={} not found.'.format(pid)})
    ```

    :param int status: HTTP status code to send back. Default 500
    :param error_data: Error data to send back
    """
    # NOTE: adjusted to output json format 
    # print traceback to logger if the exception is passed in
    # we don't need to log every case that comes through here since
    # some exception are expected and handle properly outside this function
    if ex is not None:
        current_app.logger.error(traceback.format_exc())
    payload = {}
    payload['status'] = status

    if status == 500 and error_data is None:
        payload['message'] = 'Something went wrong while processing this.'
    return payload


def test_print_helper(*args):
    pos = 1
    failed = False
    msg = []
    for arg in args:
        if not arg:
            failed = True
            msg.append(f"Test failed, with condition v{pos} is {arg}")
        pos += 1
    return (not failed, msg)


class CSGServerError(Exception):
    def __init__(self, error_message, response_code, response_body):

        Exception.__init__(self, error_message)
        self.response_code = response_code # Http status code
        self.response_body = response_body # Full http response
        
        # Parsed error message from server
        try:
            # if we receive str/bytes we try to convert to unicode/str to have
            # consistent message types (see #616)
            if (isinstance(error_message, bytes)):
                self.error_message = error_message.decode()
            else:
                self.error_message = error_message
        except Exception:
            self.error_message = error_message

    def __str__(self):
        if self.response_code is not None:
            return f"{self.response_code}: {self.error_message}"
        else:
            return f"{self.error_message}"


def on_http_error(error):
    """Manage GitlabHttpError exceptions.
    This decorator function can be used to catch GitlabHttpError exceptions
    raise specialized exceptions instead.
    Args:
        The exception type to raise -- must inherit from GitlabError
    """

    def wrap(f: __F) -> __F:
        @functools.wraps(f)
        def wrapped_f(*args: Any, **kwargs: Any) -> Any:
            try:
                return f(*args, **kwargs)
            except GitlabHttpError as e:
                raise error(e.error_message, e.response_code, e.response_body) from e

        return cast(__F, wrapped_f)

    return wrap